# koa 框架

## 复习

### (1) nodejs作用

1. 写服务接口
2. 操作本地的文件

### (2) 服务器

* 使用nodejs启动服务
* live-server也可以启动服务
* tomcat可以启动服务
* *Nginx*

## nodejs相关框架

* express.js框架
* koa.js框架(express框架的原班人马开发)
* egg.js框架(阿里)
* hapi.js框架(沃尔玛)
* nestjs(对ts的支持)

## (一) hello word`*`

1. 初始化package.json

   新建目录koademo，执行 npm init，一路回车

2. 创建一个简单的koa应用（hello world）

   安装koa模块 npm i koa --save
   在项目根目录新建app.js，app.js代码如下：

   ```javascript
   // 导入koa
   const Koa = require('koa');
   // 创建对象app
   const app = new Koa();
   
   app.use(function(ctx) {
       // 给前端返回数据
       ctx.body = 'hello world';
   });
   
   // 设置服务器端口
   app.listen(3000,()=> {
       console.log('http://localhost:3000');
   })
   ```

3. 执行node app.js
4. 用浏览器访问 http://localhost:3000
## (二) 服务器自动重新部署

我们发现没修改一次代码，都必须重启才能生效，这显然不方便，我们希望修改了文件服务能自动重启.nodemon能实现这样功能

1. 全局安装npm i nodemon -g 
2. 启动服务的时候用nodemon app.js 代替node app.js

## (三) koa路由 koa-router`*`

1. 创建新文件夹koa-app作为项目的跟目录

2. 使用终端(命令提示符或者powershell)进入根目录,初始化package.json   

   ```
   npm init -y
   ```

3. 安装koa和koa-router 

   ```
   npm i koa koa-router --save-dev
   ```

4. 新建app.js,代码如下

   ```typescript
   // 导入koa
   const Koa = require('koa');
   // 导入koa-router
   const Router = require('koa-router');
   // 创建对象app
   const app = new Koa();
   // 创建一个router对象, 用来配置路由
   const router = new Router();
   
   // 配置接口: 第一个参数是接口地址, 第二个参是个函数, 在里面返回前端需要的数据
   // ctx包含请求返回数据等相关的东西
   router.all('/',ctx=> {
       ctx.body = `
       <ol>
           <li><a href="/add">新增</a></li>
           <li><a href="/list">列表</a></li>
           <li><a href="/del">删除</a></li> 
       </ol>
       `;
   }); 
   
   router.all('/add', ctx=> {
       ctx.body = '新增';
   });
   
   router.all('/list', ctx=> {
       // ctx.body = '列表';
       ctx.body = ['a','b','c'];
   });
   
   router.all('/del', ctx=> {
       ctx.body = '删除';
   });
   
   // 使用router
   app.use(router.routes());
    
   
   // 设置服务器端口
   app.listen(3000,()=> {
       console.log('http://localhost:3000');
   })
   ```
   
5. 启动服务，用浏览器访接口即可

6. 封装router,我们一般会把router单独放到一个文件,然后倒入到app.js使用

   在根目录新建文件夹router, 在其中新建filmRouter.js和cinemaRouter.js
   
   filmRouter.js代码如下, cinemaRouter.js代码(略)
   
   ```typescript
   /**
    * 电影模块的路由
    */
   const Router = require('koa-router');
   const router = new Router();
   
   // 配置接口: 第一个参数是接口地址, 第二个参是个函数, 在里面返回前端需要的数据
   // ctx包含请求返回数据等相关的东西
   router.all('/film',ctx=> {
       ctx.body = `
       <ol>
           <li><a href="/film/add">新增电影</a></li>
           <li><a href="/film/list">电影列表</a></li>
           <li><a href="/film/del">删除电影</a></li> 
       </ol>
       `;
   }); 
   
   router.all('/film/add', ctx=> {
       ctx.body = '新增电影';
   });
   
   router.all('/film/list', ctx=> {
       // ctx.body = '列表';
       ctx.body = ['a','b','c'];
   });
   
   router.all('/film/del', ctx=> {
       ctx.body = '删除电影';
   });  
   // 把router导出
   module.exports = router;
   ```
   
   app.js里的代码变成这样
   
   ```typescript
   // 导入koa
   const Koa = require("koa");
   // 导入koa-router
   const Router = require("koa-router");
   // 创建对象app
   const app = new Koa();
   const router = new Router();
   router.all("/", (ctx) => {
     ctx.body = `
       <ol>
           <li><a href="/film">电影模块</a></li>
           <li><a href="/cinema">影院模块</a></li>
       </ol> 
       `;
   });
   
   // 导入电影路由
   const filmRouter = require("./router/filmRouter");
   const cinemaRouter = require("./router/cinemaRouter"); 
   
   // 使用路由
   app.use(router.routes());
   app.use(filmRouter.routes());
   app.use(cinemaRouter.routes());
   
   // 设置服务器端口
   app.listen(3000, () => {
     console.log("http://localhost:3000");
   });
   
   ```
   
   浏览器访问http://localhost:3000

## (四) koa中间件`*`

一个请求从发出到返回数据,如果我们想在这个过程里做一些统一的操作,可以使用中间件来完成.

1. 中间件是个函数

2. 使用中间件的方式, app.use(中间件)

   ```typescript
   app.use((ctx, next) => {
   	// 在ctx上放入username,后面的所有请求的ctx里都会有username这个变量
     ctx.username = '我是老胡';
     // 处理完之后放行,不使用next()的话,程序会被挂起来不动了
     next();  // 在vue路由守卫里曾使用过next
   })
   ```

3. 中间件是有顺序的

4. 中间的运行顺序

   跟你去北京的双程路线一样, 执行顺序是135(去)  => 642(回)
   
   ```javascript
   const Koa = require('koa');
   const app = new Koa(); 
   
   app.use((ctx,next)=> { 
       console.log(1);
       ctx.body = 'hello';
       next();
       console.log(2);
   })
   app.use((ctx,next)=> {
       console.log(3); 
       next();
       console.log(4);
   })
   app.use((ctx,next)=> {
       console.log(5);
       next();
       console.log(6);
   })
   
   app.listen(3000,()=> {
       console.log('http://localhost:3000');
   });
   ```
   
   

## (五) 设置静态目录`*`

1. 在目录中创建目录static，在public下创建文件demo.html，访问http://localhost:3000/public/demo.html是无法访问得到，因为我们还没有设置静态资源目录,设置静态资源目录要用到koa-static模块

2. 安装koa-static 

   ```
   npm i koa-static --save-dev
   ```

3. 在app.j是里加入如下代码

```
const koaStatic = require('koa-static');
app.use(koaStatic(__dirname + '/public'));
```

再来访问 http://localhost:3000/demo.html 就可以访问了,ps: 路径不用加public

## (六) 获取请求参数`*`

* 获取get请求的参数, 使用ctx.query
* 获取请求, 先安装和使用koa-body, 然后通过ctx.request.body来获取



给刚才的demo.html添加axios用来发送请求，代码如下：
```typescript
<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <meta http-equiv="X-UA-Compatible" content="ie=edge">
  <title>Document</title>
  <script src="https://cdn.bootcss.com/axios/0.19.0-beta.1/axios.js"></script>
</head>
<body>
  <h3>
    测试koa静态资源目录
  </h3>

  <script>
    let url = 'http://localhost:3000/add';
    let data = {
      params: {
        username: 'laohu',
        phone: 13800000000
      }
    }
    axios.get(url, data).then(res => {
      console.log(res);
    })

  </script>
</body>
</html>

```
1. 获取get请求参数
在router.js的add接口里加入如下代码
```typescript
router.all('/add', ctx => {
  // 获取get请求参数
  const username = ctx.query.username;
  const phone = ctx.query.phone;
  // 把拿到的数据放入ctx.body
  ctx.body = {
    module: 'add',
    username,
    phone
  }
})
```

2. 获取post请求
   获取post请求需要使用koa-body模块
   安装koa-body npm i koa-body --save

   ```
   npm i koa-body --save
   ```

   在app.js里加入如下代码：
   
   ```JavaScript
   const koaBody = require('koa-body');
   app.use(koaBody());
   ```
   
   获取post请求参数的代码如下
   
   ```JavaScript
   ctx.request.body.xxx
   ```

3. 使用中间件封装请求参数

   ```typescript
   // 把get请求参数和post请求参数都放入params对象
   app.use(async(ctx, next) => {
     ctx.params = {
       ...ctx.query,
       ...ctx.request.body
     }
     await next();
   });
   ```


## (七) 使用模板(了解)
一般请求一个接口返回的是一坨数据，然而有时候我们希望返回的是一个html网页或者一段html代码（上周分享的服务器渲染）
我们试用koa-swig模块来向前端返回一个html

1. 安装koa-swig 

```
npm i koa-view --save-dev
```

在根目录创建views目录，在views目录下创建tpl.html,代码如下

```typescript
<!DOCTYPE html>
<html lang="en">

<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <meta http-equiv="X-UA-Compatible" content="ie=edge">
  <title>Document</title>
</head>

<body>
  <h3><%=title %></h3>
  <ul>
    <%
      list.forEach(item=>{
    %>
    <li><%=item.username %> => <%=item.age %></li>
    <%
    })
    %>
  </ul>
</body> 
</html>
```

2. 在app.js添加如下代码：

```typescript
const path = require('path');
const views = require("koa-views");
app.use(views(path.join(__dirname, "views"), {
  map: {
    html: 'underscore'
  }
}));
   
// 在路由中使用   
router.all('/', async ctx => {
  const list = [{
      username: '约书亚',
      age: 22
    },
    {
      username: '亚伯拉罕',
      age: 100
    },
    {
      username: '挪亚',
      age: 500
    }
  ] 

  await ctx.render('tpl', {
    title: 'kao-框架实战',
    list
  });
})
```
3. 访问 http://localhost:3000，就可以看到一个html页面

ps: 要访问一个接口,吐出一个页面你还可以直接在js里使用反引号``拼接模板,当然也可以使用vue来渲染

## (八) 跨域配置`*`

自己编写一个中间件即可,代码如下

```typescript
app.use(async(ctx, next) => {
  ctx.set("Access-Control-Allow-Origin", "*");
  ctx.set("Access-Control-Allow-Methods", "PUT, POST, GET, DELETE, OPTIONS");
  // 请求头设置
  ctx.set(
    "Access-Control-Allow-Headers",
    `Content-Type, Content-Length, Authorization, Accept, X-Requested-With , yourHeaderFeild,x-token,sessionToken,token`
  );
  if (ctx.method == "OPTIONS") {
    ctx.body = 200;
  } else {
    await next();
  }
})
```

访问刚才的demo.html文件, http://localhost:3000/demo.html,你会发现你的请求的响应头上添加了Access-Control-Allow-Origin: *,说明服务器跨域设置已经成功了.

## (九) 图片上传`*`

content-type请求数据的时候, 设置content-type来告诉后端我们发送过去的数据格式是什么样的

三个常用的content-type

* application/json  默认
  ![image-20220927101901013](../../docs/img/image-20220927101901013.png)
* application/x-www-form-urlencoded post请求
  ![image-20220927101957450](bugImg/image-20220927101957450.png)

* multipart/form-data 文件上传

  ![image-20220927103316765](../../docs/img/image-20220927103316765.png)

app.js

```javascript
const koaBody = require("koa-body");

 app.use(koaBody({
  multipart: true,
  formidable: {
    //上传文件存储目录
    uploadDir: path.join(__dirname, `/public/upload/`),
    //允许保留后缀名
    keepExtensions: true,
    multipart: true,
  },
  jsonLimit: '10mb',
  formLimit: '10mb',
  textLimit: '10mb'
})); //解析formdata过来的数据 
```

uploadRouter.js, 同时需要在app.js中去加载和使用

```javascript
const Router = require("koa-router"); 
const router = new Router(); 

// 上传一张图片
router.all('/upload/one', async ctx=> {
    console.log(ctx.request.files.file.filepath)
    let imgName = ctx.request.files.file.filepath.split("\\upload\\")[1];
    ctx.body = {
        code: 666,
        msg:'上传成功',
        data: {
            imgUrl: 'http://localhost:3000/upload/'+imgName
        }
    }
})

module.exports = router;
```

前端文件-html

```html
<!DOCTYPE html>
<html lang="en">

<head>
    <meta charset="UTF-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Document</title>
</head>

<body>
    <input type="file">
    <button onclick="upload()">upload</button> 
    <script src="https://cdn.bootcdn.net/ajax/libs/axios/0.27.2/axios.min.js"></script>
    <script>
       function upload() {
         // todo
       }
    </script>
</body>

</html>
```



##  (十) websocket

app.js

```javascript
// 导入WebSocket模块:
const WebSocket = require('ws'); // npm i ws
// 引用Server类:
const WebSocketServer = WebSocket.Server;

// 实例化:
const wss = new WebSocketServer({
    port: 3000
});

wss.on('connection', function (ws) {
    ws.on('message', function (message) {
        console.log(message.toString());
        ws.send('你发的消息:' + message + ',我收到了');
    })
});
```

浏览器端

```html
<body>

    <input type="text" onblur="send()">
    <script>
        // 打开一个WebSocket:
        var ws = new WebSocket('ws://localhost:3000');
        // 响应onmessage事件:
        ws.onmessage = function (msg) {
            console.log(msg.data);
        };
        ws.onopen = function () {
            // 给服务器发送一个字符串:
            ws.send('Hello!');
        }

        function send() {
            var value = event.currentTarget.value;
            ws.send(value);
        }
    </script>
</body>
```



## 

